unit Filt_tbl;

{ Version 1.0 Apr-27-1997

  TFilterTable is a TTable-descendant for Delphi 1 that has some
  properties to filter records that are compatible to the TTable
  of Delphi 2.

  TFilterTable does nothing in Delphi 2, but if you add it to the
  component-palette, you can use (more or less) the same sourcecode
  for both Win16 and Win32 projects.
}


{$IFNDEF WIN32}
  {$DEFINE FilterTable} { if not defined, you wil get a dummy }
  {$DEFINE CopyTable}   { used for a save OnFilterRecord-event }
{$ENDIF NDEF WIN32}

{
  TFilterExpr, TExprParser:
    Copyright (c) 1995,96 Borland International
    I have copied the Delphi 2 sourcecode and adapted it to 16-bit.
  TFilterTable:
    (C) 1997 Christoph R. Kirchner
    The idea with FSameTable (a PRIVATE SECRET table, used for a
    save OnFilterRecord-event) came from a source I have never found
    again. Please tell me if you know wich author I can thank for
    this idea.

  Freeware, the author disclaims all warranties (see Readme.txt).

  ckirchner@geocities.com

}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Controls,
  Forms, Dialogs, DB, DBTables, DbiErrs, DBITypes, DBIProcs;

const
{ RegisterPage = 'Datenzugriff'; { german }
  RegisterPage = 'Data Access'; {}

type

{$IFDEF FilterTable}
  TFilterRecordEvent = procedure(DataSet: TDataSet;
    var Accept: Boolean) of object;
  TFilterOption = (foCaseInsensitive, foNoPartialCompare);
  TFilterOptions = set of TFilterOption;

  TFilterTable = Class(TTable)
  private
  {$IFDEF CopyTable }
    FSameTable: tTable; { a PRIVATE SECRET table. }
  {$ENDIF DEF CopyTable }
    FFilterBuffer: PChar;
    FFilterText: string;
    FFilterOptions: TFilterOptions;
    FExprFilter: HDBIFilter;
    FFuncFilter: HDBIFilter;
    FOnFilterRecord: TFilterRecordEvent;
    FFiltered: Boolean;
    procedure SetFilterData(const Text: string; Options: TFilterOptions);
    procedure SetFilterText(const Value: string);
    procedure SetFilterOptions(Value: TFilterOptions);
    procedure SetOnFilterRecord(const Value: TFilterRecordEvent);
    procedure SetFiltered(Value: Boolean);
    procedure ActivateFilters;
    procedure DeactivateFilters;
    function CreateExprFilter(const Expr: string;
      Options: TFilterOptions; Priority: Integer): HDBIFilter;
    function CreateFuncFilter(FilterFunc: Pointer;
      Priority: Integer): HDBIFilter;
(*  FilterFunc: Pointer to:
      pfGENFilter = function (
        ulClientData  : Longint; { used for "self" }
        pRecBuf       : Pointer;
        iPhyRecNum    : Longint
      ): Integer; *)
    procedure SetFilterHandle(var Filter: HDBIFilter; Value: HDBIFilter);
{   function RecordFilter(RecBuf: Pointer; RecNo: Longint): Smallint; {stdcall;}
  protected
    procedure DoAfterOpen; override;
    procedure DoAfterClose; override;
  public
    Constructor Create(AOwner: TComponent); override;
  published
    property Filter: string read FFilterText write SetFilterText;
    property Filtered: Boolean read FFiltered write SetFiltered default False;
    property FilterOptions: TFilterOptions read FFilterOptions write SetFilterOptions default [];
    property OnFilterRecord: TFilterRecordEvent read FOnFilterRecord write SetOnFilterRecord;
  End;

{$ELSE DEF FilterTable}
  TFilterTable = Class(TTable)
  end;
{$ENDIF DEF FilterTable}

Procedure Register;


implementation

{$IFDEF WIN32}
  {$R Filt_Tbl.d32} { 32-bit DCR is renamed to d32 }
{$ELSE DEF WIN32}
  {$R Filt_Tbl.d16} { 16-bit DCR is renamed to d16 }
{$ENDIF DEF WIN32}


{$IFDEF FilterTable}
{$P+}

{ TFilterExpr }

const
  Unassigned = '';

type

  Variant = String; { Variant does not exist in Delphi 1 }

  ExprBadCompare = class(Exception);

  TExprNodeKind = (enField, enConst, enOperator);

  PExprNode = ^TExprNode;
  TExprNode = record
    FNext: PExprNode;
    FKind: TExprNodeKind;
    FPartial: Boolean;
    FOperator: CanOp;
    FData: Variant;
    FLeft: PExprNode;
    FRight: PExprNode;
  end;

  TFilterExpr = class(TObject)
  private
    FDataSet: TDataSet;
    FOptions: TFilterOptions;
    FNodes: PExprNode;
    FExprBuffer: PCANExpr;
    FExprBufSize: Integer;
    FExprNodeSize: Integer;
    FExprDataSize: Integer;
    function FieldFromNode(Node: PExprNode): TField;
    function GetExprData(Pos, Size: Integer): PChar;
    function PutCompareNode(Node: PExprNode): Integer;
    function PutConstBCD(const Value: Variant; Decimals: Integer): Integer;
    function PutConstDate(const Value: Variant): Integer;
    function PutConstDateTime(const Value: Variant): Integer;
    function PutConstFloat(const Value: Variant): Integer;
    function PutConstInt(DataType: Integer; const Value: Variant): Integer;
    function PutConstNode(DataType: Integer; Data: PChar;
      Size: Integer): Integer;
    function PutConstStr(const Value: string): Integer;
    function PutConstTime(const Value: Variant): Integer;
    function PutData(Data: PChar; Size: Integer): Integer;
    function PutExprNode(Node: PExprNode): Integer;
    function PutFieldNode(Field: TField): Integer;
    function PutNode(NodeType: NodeClass; OpType: CanOp;
      OpCount: Integer): Integer;
    procedure SetNodeOp(Node, Index, Data: Integer);
  public
    constructor Create(DataSet: TDataSet; Options: TFilterOptions);
    destructor Destroy; override;
    function NewCompareNode(Field: TField; Operator: CanOp;
      const Value: Variant): PExprNode;
    function NewNode(Kind: TExprNodeKind; Operator: CanOp;
      const Data: Variant; Left, Right: PExprNode): PExprNode;
    function GetFilterData(Root: PExprNode): PCANExpr;
  end;

constructor TFilterExpr.Create(DataSet: TDataSet; Options: TFilterOptions);
begin
  inherited Create;
  FDataSet := DataSet;
  FOptions := Options;
end;

destructor TFilterExpr.Destroy;
var
  Node: PExprNode;
begin
  FreeMem(FExprBuffer, FExprBufSize);
  while FNodes <> nil do
  begin
    Node := FNodes;
    FNodes := Node^.FNext;
    Dispose(Node);
  end;
  inherited Destroy;
end;

function TFilterExpr.FieldFromNode(Node: PExprNode): TField;
begin
  Result := FDataSet.FieldByName(Node^.FData);
{  if Result.FieldKind <> fkData then
    DBErrorFmt(SExprBadField, [Result.FieldName]); }
end;

function TFilterExpr.GetExprData(Pos, Size: Integer): PChar;
var
  p: PChar;
begin
{
  ReallocMem(FExprBuffer, FExprBufSize, FExprBufSize + Size);
  Move(PChar(FExprBuffer)[Pos], PChar(FExprBuffer)[Pos + Size],
    FExprBufSize - Pos);
}
  p := AllocMem(FExprBufSize + Size);
  Move(FExprBuffer^, p^, FExprBufSize);
  FreeMem(FExprBuffer, FExprBufSize);
  PChar(FExprBuffer) := p;

  Move(PChar(FExprBuffer)[Pos], PChar(FExprBuffer)[Pos + Size],
    FExprBufSize - Pos);
  Inc(FExprBufSize, Size);
  Result := PChar(FExprBuffer) + Pos;
end;

function TFilterExpr.GetFilterData(Root: PExprNode): PCANExpr;
begin
  FExprBufSize := SizeOf(CANExpr);
  FExprBuffer := AllocMem(FExprBufSize);
  PutExprNode(Root);
  with FExprBuffer^ do
  begin
    iVer := CANEXPRVERSION;
    iTotalSize := FExprBufSize;
    iNodes := $FFFF;
    iNodeStart := SizeOf(CANExpr);
    iLiteralStart := FExprNodeSize + SizeOf(CANExpr);
  end;
  Result := FExprBuffer;
end;

function TFilterExpr.NewCompareNode(Field: TField; Operator: CanOp;
  const Value: Variant): PExprNode;
begin
  Result := NewNode(enOperator, Operator, Unassigned,
    NewNode(enField, canNOTDEFINED, Field.FieldName, nil, nil),
    NewNode(enConst, canNOTDEFINED, Value, nil, nil));
end;

function TFilterExpr.NewNode(Kind: TExprNodeKind; Operator: CanOp;
  const Data: Variant; Left, Right: PExprNode): PExprNode;
begin
  New(Result);
  with Result^ do
  begin
    FNext := FNodes;
    FKind := Kind;
{   FPartial := False; }
    FPartial := True; { changed by CK }
    FOperator := Operator;
    FData := Data;
    FLeft := Left;
    FRight := Right;
  end;
  FNodes := Result;
end;

function TFilterExpr.PutCompareNode(Node: PExprNode): Integer;
const
  ReverseOperator: array[canEQ..canLE] of CanOp = (
    canEQ, canNE, canLT, canGT, canLE, canGE);
var
  Operator: CanOp;
  Left, Right, Temp: PExprNode;
  Field: TField;
  FieldPos, ConstPos, CaseInsensitive, PartialLength, L: Integer;
  S: string;
begin
  Operator := Node^.FOperator;
  Left := Node^.FLeft;
  Right := Node^.FRight;
  if Right^.FKind = enField then
  begin
    Temp := Left;
    Left := Right;
    Right := Temp;
    Operator := ReverseOperator[Operator];
  end;
  if (Left^.FKind <> enField) or (Right^.FKind <> enConst) then
    raise ExprBadCompare.Create('error in PutCompareNode: BadCompare');
  Field := FieldFromNode(Left);
  if (Right^.FData = '') then
  begin
    case Operator of
      canEQ: Operator := canISBLANK;
      canNE: Operator := canNOTBLANK;
    else
      raise ExprBadCompare.Create('error in PutCompareNode: BadNullTest');
    end;
    Result := PutNode(nodeUNARY, Operator, 1);
    SetNodeOp(Result, 0, PutFieldNode(Field));
  end else
  begin
    if ((Operator = canEQ) or (Operator = canNE)) and
      (Field.DataType = ftString) then
    begin
      S := Right^.FData;
      L := Length(S);
      if L <> 0 then
      begin
        CaseInsensitive := 0;
        PartialLength := 0;
        if foCaseInsensitive in FOptions then CaseInsensitive := 1;
        if Node^.FPartial then PartialLength := L else
          if not (foNoPartialCompare in FOptions) and (L > 1) and
            (S[L] = '*') then
          begin
            Delete(S, L, 1);
            PartialLength := L - 1;
          end;
        if (CaseInsensitive <> 0) or (PartialLength <> 0) then
        begin
          Result := PutNode(nodeCOMPARE, Operator, 4);
          SetNodeOp(Result, 0, CaseInsensitive);
          SetNodeOp(Result, 1, PartialLength);
          SetNodeOp(Result, 2, PutFieldNode(Field));
          SetNodeOp(Result, 3, PutConstStr(S));
          Exit;
        end;
      end;
    end;
    Result := PutNode(nodeBINARY, Operator, 2);
    FieldPos := PutFieldNode(Field);
    case Field.DataType of
      ftString:
        ConstPos := PutConstStr(Right^.FData);
      ftSmallint:
        ConstPos := PutConstInt(fldINT16, Right^.FData);
      ftInteger:
        ConstPos := PutConstInt(fldINT32, Right^.FData);
      ftWord:
        ConstPos := PutConstInt(fldUINT16, Right^.FData);
      ftFloat, ftCurrency:
        ConstPos := PutConstFloat(Right^.FData);
      ftBCD:
        ConstPos := PutConstBCD(Right^.FData, Field.Size);
      ftDate:
        ConstPos := PutConstDate(Right^.FData);
      ftTime:
        ConstPos := PutConstTime(Right^.FData);
      ftDateTime:
        ConstPos := PutConstDateTime(Right^.FData);
    else
      raise ExprBadCompare.Create('error in PutCompareNode: BadField ' +
                                  Field.FieldName);
    end;
    SetNodeOp(Result, 0, FieldPos);
    SetNodeOp(Result, 1, ConstPos);
  end;
end;

type
  Currency = Double;
(*
  pFMTBcd = ^FMTBcd;
  FMTBcd  = packed record
    iPrecision: Byte;         { 1..64 considered valid }
    iSignSpecialPlaces: Byte; { sign:1, special:1, places:6 }
  { bcd nibbles, 00..99 per byte, high nibble 1st: }
    iFraction: packed array [0..31] of Byte;
  end;
function DbiBcdFromFloat (      { Converts FLOAT number into FMTBcd format }
var   iVal          : Double;   { Float to convert }
      iPrecision    : Word;     { Precision of BCD }
      iPlaces       : Word;     { Number of decimals }
var   Bcd           : FMTBcd    { returns Bcd number (length = iPrecision +2) }
   ): DBIResult stdcall;
*)
function TFilterExpr.PutConstBCD(const Value: Variant;
  Decimals: Integer): Integer;
var
  C: Currency;
  BCD: FMTBcd;
begin
  C := StrToFloat(Value);
  DbiBcdFromFloat(C, 32, Decimals, BCD);
  Result := PutConstNode(fldBCD, @BCD, 18);
end;

const
{ Seconds and milliseconds per day }
  SecsPerDay = 24 * 60 * 60;
  MSecsPerDay = SecsPerDay * 1000;

type
  TTimeStamp = record
    Time: Longint;      { Number of milliseconds since midnight }
    Date: Longint;      { One plus number of days since 1/1/0001 }
  end;

function DateTimeToTimeStamp(DateTime: TDateTime): TTimeStamp;
var
  FracDateTime: Single;
begin
  result.Date := Trunc(DateTime);
  FracDateTime := Frac(DateTime);
  result.Time := Round(MSecsPerDay * FracDateTime);
end;

function TimeStampToMSecs(const TimeStamp: TTimeStamp): Comp;
begin
  result :=  TimeStamp.Date;
  result :=  result * MSecsPerDay + TimeStamp.Time;
end;

function TFilterExpr.PutConstDate(const Value: Variant): Integer;
var
  DateTime: TDateTime;
  TimeStamp: TTimeStamp;
begin
  DateTime := StrToDateTime(Value);
  TimeStamp := DateTimeToTimeStamp(DateTime);
  Result := PutConstNode(fldDATE, @TimeStamp.Date, 4);
end;

function TFilterExpr.PutConstDateTime(const Value: Variant): Integer;
var
  DateTime: TDateTime;
  TimeStamp: TTimeStamp;
  DateData: Double;
begin
  DateTime := StrToDateTime(Value);
  TimeStamp := DateTimeToTimeStamp(DateTime);
  DateData := TimeStampToMSecs(TimeStamp);
  Result := PutConstNode(fldTIMESTAMP, PChar(@DateData), 8);
end;

function TFilterExpr.PutConstTime(const Value: Variant): Integer;
var
  DateTime: TDateTime;
  TimeStamp: TTimeStamp;
begin
  DateTime := StrToTime(Value);
  TimeStamp := DateTimeToTimeStamp(DateTime);
  Result := PutConstNode(fldTIME, @TimeStamp.Time, 4); 
end;

function TFilterExpr.PutConstFloat(const Value: Variant): Integer;
var
  F: Double;
begin
  F := StrToFloat(Value);
  Result := PutConstNode(fldFLOAT, PChar(@F), SizeOf(Double));
end;

function TFilterExpr.PutConstInt(DataType: Integer;
  const Value: Variant): Integer;
var
  I: Longint;
  Size: Integer;
begin
  I := StrToInt(Value);
  Size := 2;
  case DataType of
    fldINT16:
      if (I < -32768) or (I > 32767) then
        raise ExprBadCompare.Create('PutConstInt fldINT16: RangeError');
    fldUINT16:
      if (I < 0) or (I > 65535) then
        raise ExprBadCompare.Create('PutConstInt fldUINT16: RangeError');
  else
    Size := 4;
  end;
  Result := PutConstNode(DataType, PChar(@I), Size);
end;

function TFilterExpr.PutConstNode(DataType: Integer; Data: PChar;
  Size: Integer): Integer;
begin
  Result := PutNode(nodeCONST, canCONST2, 3);
  SetNodeOp(Result, 0, DataType);
  SetNodeOp(Result, 1, Size);
  SetNodeOp(Result, 2, PutData(Data, Size));
end;

function TFilterExpr.PutConstStr(const Value: string): Integer;
var
  Buffer: array[0..255] of Char;
begin
  AnsiToNative(FDataSet.Locale, Value, Buffer, SizeOf(Buffer) - 1);
  Result := PutConstNode(fldZSTRING, Buffer, StrLen(Buffer) + 1);
end;

function TFilterExpr.PutData(Data: PChar; Size: Integer): Integer;
begin
  Move(Data^, GetExprData(FExprBufSize, Size)^, Size);
  Result := FExprDataSize;
  Inc(FExprDataSize, Size);
end;

function TFilterExpr.PutExprNode(Node: PExprNode): Integer;
const
  BoolFalse: WordBool = False;
var
  Field: TField;
begin
  case Node^.FKind of
    enField:
      begin
        Field := FieldFromNode(Node);
        if Field.DataType <> ftBoolean then
          raise ExprBadCompare.Create('Not Boolean: ' + Field.FieldName);
        Result := PutNode(nodeBINARY, canNE, 2);
        SetNodeOp(Result, 0, PutFieldNode(Field));
        SetNodeOp(Result, 1, PutConstNode(fldBOOL, PChar(@BoolFalse),
          SizeOf(WordBool)));
      end;
    enOperator:
      case Node^.FOperator of
        canEQ..canLE:
          Result := PutCompareNode(Node);
        canAND, canOR:
          begin
            Result := PutNode(nodeBINARY, Node^.FOperator, 2);
            SetNodeOp(Result, 0, PutExprNode(Node^.FLeft));
            SetNodeOp(Result, 1, PutExprNode(Node^.FRight));
          end;
      else
        Result := PutNode(nodeUNARY, canNOT, 1);
        SetNodeOp(Result, 0, PutExprNode(Node^.FLeft));
      end;
  else
    raise ExprBadCompare.Create('Expr. Incorrect');
  end;
end;

function TFilterExpr.PutFieldNode(Field: TField): Integer;
var
  Buffer: array[0..255] of Char;
begin
  AnsiToNative(FDataSet.Locale, Field.FieldName, Buffer, SizeOf(Buffer) - 1);
  Result := PutNode(nodeFIELD, canFIELD2, 2);
  SetNodeOp(Result, 0, Field.FieldNo);
  SetNodeOp(Result, 1, PutData(Buffer, StrLen(Buffer) + 1));
end;

function TFilterExpr.PutNode(NodeType: NodeClass; OpType: CanOp;
  OpCount: Integer): Integer;
var
  Size: Integer;
begin
  Size := SizeOf(CANHdr) + OpCount * SizeOf(Word);
  with PCANHdr(GetExprData(SizeOf(CANExpr) + FExprNodeSize, Size))^ do
  begin
    nodeClass := NodeType;
    canOp := OpType;
  end;
  Result := FExprNodeSize;
  Inc(FExprNodeSize, Size);
end;

procedure TFilterExpr.SetNodeOp(Node, Index, Data: Integer);
begin
  PWordArray(PChar(FExprBuffer) + (SizeOf(CANExpr) + Node +
    SizeOf(CANHdr)))^[Index] := Data;
end;

{ TExprParser }

type

  TExprToken = (etEnd, etSymbol, etName, etLiteral, etLParen, etRParen,
    etEQ, etNE, etGE, etLE, etGT, etLT);

  TExprParser = class
  private
    FFilter: TFilterExpr;
    FText: string;
    FSourcePtr: PChar;
    FTokenPtr: PChar;
    FTokenString: array[0..255] of char;
    FToken: TExprToken;
    FFilterData: PCANExpr;
    procedure NextToken;
    function ParseExpr: PExprNode;
    function ParseExpr2: PExprNode;
    function ParseExpr3: PExprNode;
    function ParseExpr4: PExprNode;
    function ParseExpr5: PExprNode;
    function TokenName: string;
    function TokenSymbolIs(const S: string): Boolean;
  public
    constructor Create(DataSet: TDataSet; const Text: string;
      Options: TFilterOptions);
    destructor Destroy; override;
    property FilterData: PCANExpr read FFilterData;
  end;

constructor TExprParser.Create(DataSet: TDataSet; const Text: string;
  Options: TFilterOptions);
var
  Root: PExprNode;
  pc: array[0..255] of Char;
begin
  FFilter := TFilterExpr.Create(DataSet, Options);
  FText := Text;
  FSourcePtr := StrPCopy(pc, Text);
  NextToken;
  Root := ParseExpr;
  if FToken <> etEnd then
    raise ExprBadCompare.Create('Expr. Termination');
  FFilterData := FFilter.GetFilterData(Root);
end;

destructor TExprParser.Destroy;
begin
  FFilter.Free;
end;

procedure TExprParser.NextToken;
var
  P, TokenStart: PChar;
  L: Integer;
  StrBuf: array[0..255] of Char;
begin
  FTokenString[0] := #0;
  P := FSourcePtr;
  while (P^ <> #0) and (P^ <= ' ') do Inc(P);
  FTokenPtr := P;
  case P^ of
    'A'..'Z', 'a'..'z', '_':
      begin
        TokenStart := P;
        Inc(P);
        while P^ in ['A'..'Z', 'a'..'z', '0'..'9', '_'] do Inc(P);
        StrLCopy(FTokenString, TokenStart, P - TokenStart);
        FToken := etSymbol;
      end;
    '[':
      begin
        Inc(P);
        TokenStart := P;
        while (P^ <> ']') and (P^ <> #0) do Inc(P);
        if P^ = #0 then
          raise ExprBadCompare.Create('Expr. NameError');
        StrLCopy(FTokenString, TokenStart, P - TokenStart);
        FToken := etName;
        Inc(P);
      end;
    '''':
      begin
        Inc(P);
        L := 0;
        while True do
        begin
          if P^ = #0 then
            raise ExprBadCompare.Create('Expr. StringError');
          if P^ = '''' then
          begin
            Inc(P);
            if P^ <> '''' then Break;
          end;
          if L < SizeOf(StrBuf) then
          begin
            StrBuf[L] := P^;
            Inc(L);
          end;
          Inc(P);
        end;
        StrLCopy(FTokenString, StrBuf, L);
        FToken := etLiteral;
      end;
    '-', '0'..'9':
      begin
        TokenStart := P;
        Inc(P);
      { Original in Delphi 2:
        while P^ in ['0'..'9', '.', 'e', 'E', '+', '-'] do Inc(P);
        I added ',', ':', '/', to handle german float, time and us date. }
        while P^ in ['0'..'9', '.', ',', ':', '/', 'e', 'E', '+', '-'] do
          Inc(P);
        StrLCopy(FTokenString, TokenStart, P - TokenStart);
        FToken := etLiteral;
      end;
    '(':
      begin
        Inc(P);
        FToken := etLParen;
      end;
    ')':
      begin
        Inc(P);
        FToken := etRParen;
      end;
    '<':
      begin
        Inc(P);
        case P^ of
          '=':
            begin
              Inc(P);
              FToken := etLE;
            end;
          '>':
            begin
              Inc(P);
              FToken := etNE;
            end;
        else
          FToken := etLT;
        end;
      end;
    '=':
      begin
        Inc(P);
        FToken := etEQ;
      end;
    '>':
      begin
        Inc(P);
        if P^ = '=' then
        begin
          Inc(P);
          FToken := etGE;
        end else
          FToken := etGT;
      end;
    #0:
      FToken := etEnd;
  else
    raise ExprBadCompare.Create('Expr. InvalidChar: ' + P^);
  end;
  FSourcePtr := P;
end;

function TExprParser.ParseExpr: PExprNode;
begin
  Result := ParseExpr2;
  while TokenSymbolIs('OR') do
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, canOR, Unassigned,
      Result, ParseExpr2);
  end;
end;

function TExprParser.ParseExpr2: PExprNode;
begin
  Result := ParseExpr3;
  while TokenSymbolIs('AND') do
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, canAND, Unassigned,
      Result, ParseExpr3);
  end;
end;

function TExprParser.ParseExpr3: PExprNode;
begin
  if TokenSymbolIs('NOT') then
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, canNOT, Unassigned,
      ParseExpr4, nil);
  end else
    Result := ParseExpr4;
end;

function TExprParser.ParseExpr4: PExprNode;
const
  Operators: array[etEQ..etLT] of CanOp = (
    canEQ, canNE, canGE, canLE, canGT, canLT);
var
  Operator: CanOp;
begin
  Result := ParseExpr5;
  if FToken in [etEQ..etLT] then
  begin
    Operator := Operators[FToken];
    NextToken;
    Result := FFilter.NewNode(enOperator, Operator, Unassigned,
      Result, ParseExpr5);
  end;
end;

function TExprParser.ParseExpr5: PExprNode;
begin
  case FToken of
    etSymbol:
      if TokenSymbolIs('NULL') then
        Result := FFilter.NewNode(enConst, canNOTDEFINED, '', nil, nil) else
        Result := FFilter.NewNode(enField, canNOTDEFINED, StrPas(FTokenString), nil, nil);
    etName:
      Result := FFilter.NewNode(enField, canNOTDEFINED, StrPas(FTokenString), nil, nil);
    etLiteral:
      Result := FFilter.NewNode(enConst, canNOTDEFINED, StrPas(FTokenString), nil, nil);
    etLParen:
      begin
        NextToken;
        Result := ParseExpr;
        if FToken <> etRParen then
          raise ExprBadCompare.Create('Expr. NoRParen: ' + TokenName);
      end;
  else
    raise ExprBadCompare.Create('Expr. Expected: ' + TokenName);
  end;
  NextToken;
end;

function TExprParser.TokenName: string;
var
  temp: Array[0..255] of char;
begin
  if FSourcePtr = FTokenPtr then Result := 'leer' else
  begin
    StrPas(StrLCopy(temp, FTokenPtr, FSourcePtr - FTokenPtr));
    Result := '''' + Result + '''';
  end;
end;

function TExprParser.TokenSymbolIs(const S: string): Boolean;
begin
  Result := (FToken = etSymbol) and (CompareText(StrPas(FTokenString), S) = 0);
end;


{ ---------------------------------------------------------------------------}

function RecordFilter(FilterTable: Pointer; RecBuf: PChar;
  RecNo: Longint): Integer;
 export;
var
  Accept: Boolean;
begin
  Accept := True;
  with TFilterTable(FilterTable) do
  begin
    FFilterBuffer := RecBuf;
{$IFDEF CopyTable }
    if not Assigned(FSameTable) then
    begin
      Result := 1;
      exit;
    end;
    if not FSameTable.Active then FSameTable.Active := TRUE;
  {
    We now want to move the record pointer in the "secret" table
    to the record that is currently being filtered in the user's
    table.  We will then pass the secret table to the user and
    they will never know the difference :)
  }
{   Check( DBISetToRecordNo(FSameTable.Handle, RecNo) ); nur dBase }
    if (FSameTable.IndexName <> IndexName) then
      FSameTable.IndexName := IndexName;
    if (FSameTable.IndexFieldNames <> IndexFieldNames) then
      FSameTable.IndexFieldNames := IndexFieldNames;
    Check( DBISetToSeqNo(FSameTable.Handle, RecNo) );
    FSameTable.ReSync([]);
    try
      OnFilterRecord(FSameTable, Accept);
    except
      Application.HandleException(TFilterTable(FilterTable));
    end;
{$ELSE DEF CopyTable }
    try
{     StrCopy(ActiveBuffer, RecBuf);
      DataEvent(deDataSetChange, 0); }
      OnFilterRecord(TFilterTable(FilterTable), Accept);
    except
      Application.HandleException(FilterTable);
    end;
{$ENDIF DEF CopyTable }
  end;
  Result := Ord(Accept);
end;

constructor TFilterTable.Create;
begin
  inherited Create(AOwner);
  FOnFilterRecord := nil;
  FExprFilter := nil;
  FFuncFilter := nil;
  FFiltered := false;
  FFilterOptions := [];
end;

procedure TFilterTable.DoAfterOpen;
begin
  inherited DoAfterOpen;
  if FFilterText <> '' then
    FExprFilter := CreateExprFilter(FFilterText, FFilterOptions, 1);
  if Assigned(FOnFilterRecord) then
    FFuncFilter := CreateFuncFilter(@RecordFilter, 2);
  if FFiltered then
  begin
    ActivateFilters;
    First;
  end;
end;

procedure TFilterTable.DoAfterClose;
begin
  inherited DoAfterClose;
  DeactivateFilters;
  FExprFilter := nil;
  FFuncFilter := nil;
end;

procedure TFilterTable.SetOnFilterRecord(const Value: TFilterRecordEvent);
var
  Filter: HDBIFilter;
begin
  if Active then
  begin
    CheckBrowseMode;
    if Assigned(FOnFilterRecord) <> Assigned(Value) then
    begin
      if Assigned(Value) then
        Filter := CreateFuncFilter(@RecordFilter, 2)
      else
        Filter := nil;
      SetFilterHandle(FFuncFilter, Filter);
    end;
    FOnFilterRecord := Value;
    if FFiltered then First;
  end else
    FOnFilterRecord := Value;
end;

procedure TFilterTable.SetFiltered(Value: Boolean);
var
  BookMark: TBookMark;
begin
  if Active then
  begin
    CheckBrowseMode;
    if FFiltered <> Value then
    begin
      BookMark := GetBookmark;
      DbiSetToBegin(Handle);
      if Value then
      begin
        ActivateFilters;
      end
      else
        DeactivateFilters;
      FFiltered := Value;
      if not EOF then
        Next;
      First;
      try
{       GotoBookMark(BookMark); { no current rec can happen but is ok }
        DbiSetToBookmark(Handle, Bookmark);
        Resync([]);  { rmExact ? }
      except end;
      FreeBookMark(BookMark);
    end;
{   First; }
  end else
    FFiltered := Value;
end;

procedure TFilterTable.ActivateFilters;
begin
  if FExprFilter <> nil then DbiActivateFilter(Handle, FExprFilter);
  if FFuncFilter <> nil then DbiActivateFilter(Handle, FFuncFilter);
{$IFDEF CopyTable }
  FSameTable := tTable.Create(nil);
  FSameTable.DisableControls;
  FSameTable.ReadOnly := TRUE;
  FSameTable.DatabaseName := DatabaseName;
  FSameTable.TableName := TableName;
  FSameTable.Active := TRUE;
{$ENDIF DEF CopyTable }
end;

procedure TFilterTable.DeactivateFilters;
begin
  if FFuncFilter <> nil then DbiDeactivateFilter(Handle, FFuncFilter);
  if FExprFilter <> nil then DbiDeactivateFilter(Handle, FExprFilter);
{$IFDEF CopyTable }
{ Release our "secret table" }
  if Assigned(FSameTable) then
  begin
    FSameTable.Active := FALSE;
    FSameTable.Free;
    FSameTable := nil;
  end;
{$ENDIF DEF CopyTable }
end;
(*
function DbiAddFilter (                 { Add a filter to the cursor }
      hCursor       : hDBICur;          { Cursor handle }
      iClientData   : Longint;          { Client supplied data      (opt) }
      iPriority     : Word;             { 1..N with 1 being highest (opt) }
      bCanAbort     : Bool;             { TRUE if pfFiltercan return ABORT (opt) }
      pcanExpr      : pCANExpr;         { Expression tree        (opt) }
      pfFilter      : pfGENFilter;      { ptr to filter function (opt) }
var   hFilter       : hDBIFilter        { Returns filter handle }
   ): DBIResult;
*)
function TFilterTable.CreateFuncFilter(FilterFunc: Pointer;
  Priority: Integer): HDBIFilter;
begin
  Check(DbiAddFilter(Handle, Longint(Self), Priority, False, nil,
    PFGENFilter(FilterFunc), Result));
end;

function TFilterTable.CreateExprFilter(const Expr: string;
  Options: TFilterOptions; Priority: Integer): HDBIFilter;
var
  Parser: TExprParser;
begin
  Parser := TExprParser.Create(Self, Expr, Options);
  try
    Check(DbiAddFilter(Handle, 0, Priority, False, Parser.FilterData,
      nil, Result));
  finally
    Parser.Free;
  end;
end;

(*
function DbiDropFilter (                { Drop a filter }
      hCursor       : hDBICur;          { Cursor handle }
      hFilter       : hDBIFilter        { Filter handle }
   ): DBIResult;
*)

procedure TFilterTable.SetFilterHandle(var Filter: HDBIFilter;
  Value: HDBIFilter);
begin
  if FFiltered then
  begin
    CursorPosChanged;
   {DestroyLookupCursor;}
    DbiSetToBegin(Handle);
    if Filter <> nil then DbiDropFilter(Handle, Filter);
    Filter := Value;
    if Filter <> nil then DbiActivateFilter(Handle, Filter);
  end else
    Filter := Value;
end;

procedure TFilterTable.SetFilterData(const Text: string; Options: TFilterOptions);
var
  Filter: HDBIFilter;
  BookMark: TBookmark;
begin
  if Active then
  begin
    CheckBrowseMode;
    if (FFilterText <> Text) or (FFilterOptions <> Options) then
    begin
      BookMark := GetBookmark;
      if Text <> '' then
        Filter := CreateExprFilter(Text, Options, 0) else
        Filter := nil;
      SetFilterHandle(FExprFilter, Filter);
      if FFiltered then
      begin
        if not EOF then
          Next;
        First;
        try
{         GotoBookMark(BookMark); { no current rec can happen but is ok }
          DbiSetToBookmark(Handle, Bookmark);
          Resync([]);  { rmExact ? }
        except end;
      end;
      FreeBookMark(BookMark);
    end;
  end;
  FFilterText := Text;
  FFilterOptions := Options;
end;

procedure TFilterTable.SetFilterText(const Value: string);
begin
  SetFilterData(Value, FFilterOptions);
end;

procedure TFilterTable.SetFilterOptions(Value: TFilterOptions);
begin
  SetFilterData(FFilterText, Value);
end;

{$ENDIF DEF FilterTable}

procedure Register;
begin
  RegisterComponents(RegisterPage, [TFilterTable]);
end;


end.
